import List

type Matrix a = [[a]]

-- All operations are to be done on Rows,Columns and Boxes

-- transpose, built-in, converts rows to columns and back.
-- We  define boxTransp, a  function to convert boxes to rows and back!

chop i [] = []
chop i l = (take i l):chop i (drop i l)

-- constructing the boxes. Boxing exactly 3 rows is easy
boxTransp3r :: Matrix a -> Matrix a
boxTransp3r m = map concat (transpose (map (chop 3) m))

-- Boxing K rows. Box 3 rows at a time and concatenate
boxTransp :: Matrix a -> Matrix a
boxTransp = concatMap boxTransp3r. (chop 3) 
	
-- a function that  checks that a list has no duplicates 
nodups :: Eq a => [a] -> Bool
nodups [] = True
nodups (x:xs) = ([] == [z | z <- xs, z == x]) &&
                  nodups xs

-- row-wise validity
validRowWise :: Eq a => Matrix a  -> Bool
validRowWise m = and (map nodups m)

-- column-wise validity
validColWise :: Eq a => Matrix a  -> Bool
validColWise m = validRowWise (transpose m)

-- box-wise validity
validBoxWise :: Eq a => Matrix a -> Bool
validBoxWise m  = validRowWise (boxTransp m)

valid :: Eq a => Matrix a  -> Bool
valid m = validRowWise m && validColWise m && validBoxWise m 

-- choices is generates an intermediate representation. 
-- Each cell is filled by the set of all possible entries
-- permitted. 

choices :: Matrix  Char -> Matrix [Char]
choices m = map (map possValues) m
  where 
   possValues ' ' = "123456789"
   possValues '.' = "123456789"
   possValues  c = [c]

-- a function that constructs the "proDuct" of a list of lists. 
proDuct :: [[a]] -> [[a]]
proDuct [] = [[]]
proDuct (x:xs) = [y:ys | y <- x, ys <- proDuct xs]

-- collapse converts the intermediate representation to the
-- list of all grids consistent with this assignment.
-- Expand each row using proDuct. Each row is a list of possibilites now.
-- possible grids is now obtained by expanding this using proDuct.

collapse :: Matrix [a] -> [Matrix a]
collapse ir = proDuct (map proDuct ir)

-- solves the Sudoko problem
solve :: Matrix Char -> [Matrix Char]
solve = filter valid . collapse  . choices


---pruning each row

pruneRowWise :: Eq a => Matrix [a] -> Matrix [a]
pruneRowWise = map pruneRow 

pruneRow :: Eq a => [[a]] -> [[a]]
pruneRow  ls  =  map (removeSingletons) ls
  where 
  singletons = filter (\x -> ((length x) == 1))  ls
  removeSingletons [x] = [x]
  removeSingletons l = l \\ (concat singletons)

-- For columns, convert to rows, prune and convert back
pruneColWise :: Eq a => Matrix [a] -> Matrix [a]
pruneColWise m  = (transpose . pruneRowWise . transpose) m

-- For boxes, convert to rows, prune and convert back
pruneBoxWise :: Eq a => Matrix [a] -> Matrix [a]
pruneBoxWise m = (boxTransp . pruneRowWise . boxTransp) m

-- Prune them in some order
prune m = (pruneBoxWise . pruneColWise . pruneRowWise) m

-- New solution.

solve' = filter valid . collapse  . prune .  choices

-- Prune repeatedly
fixprune l = if (prune l == l) then l
		else fixprune (prune l)

solve'' = filter valid . collapse . fixprune . choices

-- generate carefully

search :: Eq a =>  Matrix [a] -> [Matrix a]
search m
  | blocked m  =  []
  | complete m =  collapse  m
  | otherwise  =  [g | m' <- expand  m
                     , g  <- search (prune m')]

--- Checks if all entries are singletons
complete :: Matrix [a] -> Bool
complete =  all (all single)

single [_] = True
single _ = False

-- Checks if there is any empty entry or duplicate occurrence of a singleton
blocked :: Eq a => Matrix [a] -> Bool
blocked m =  any (any null)  m || not (safe m)

-- Uses the "polymorphic" nature of boxes
safe :: Eq a => Matrix [a]  -> Bool
safe cm =  and ((map consistent) cm) &&
           and ((map consistent) (transpose cm)) &&
           and ((map consistent) (boxTransp cm))

-- verifies that no singleton appears twice
consistent :: Eq a => [[a]] -> Bool
consistent =  nodups . filter single


expand  m  = 
        [rows1 ++ [row1 ++ [c] : row2] ++ rows2 | c <- cs]
        where
	      (rows1,row:rows2) = break (any (not . single)) m
	      (row1,cs:row2)    = break (not . single) row


solve''' = (filter valid) . search . choices

full = [[11,12,13,14,15,16,17,18,19],
	[21,22,23,24,25,26,27,28,29],
	[31,32,33,34,35,36,37,38,39],
	[41,42,43,44,45,46,47,48,49],
	[51,52,53,54,55,56,57,58,59],
	[61,62,63,64,65,66,67,68,69],
	[71,72,73,74,75,76,77,78,79],
	[81,82,83,84,85,86,87,88,89],
	[91,92,93,94,95,96,97,98,99]]

--- Some examples 

soln1 = ["123456789",
         "456789123",
         "789123456",
         "214365897",
         "365897214",
         "897214365",
         "531642978",
         ".42978531",
         ".78531642"]

soln2 = ["123456789",
         "456 89123",
         "789 2 456",
         "214 65897",
         " 65897214",
         "897214365",
         "531 42978",
         "642 78531",
         "978531642"]

easy  =  ["2    1 38",
          "        5",
          " 7   6   ",
          "       13",
          " 981  257",
          "31    8  ",
          "9  8   2 ",
          " 5  69784",
          "4  25    "]

gentle  = [" 1 42   5",
            "  2 71 39",
            "       4 ",
            "2 71    6",
            "    4    ",
            "6    74 3",
            " 7       ",
            "12 73 5  ",
            "3   82 7 "]

diabolical =  [" 9 7  86 ",
                " 31  5 2 ",
                "8 6      ",
                "  7 5   6",
                "   3 7   ",
                "5   1 7  ",
                "      1 9",
                " 2 6  35 ",
                " 54  8 7 "]

unsolvable =  ["1  9 7  3",
                " 8     7 ",
                "  9   6  ",
                "  72 94  ",
                "41     95",
                "  85 43  ",
                "  3   7  ",
                " 5     4 ",
                "2  8 6  9"]

minimal =  [" 98      ",
             "    7    ",
             "    15   ",
             "1        ",
             "   2    9",
             "   9 6 82",
             "       3 ",
             "5 1      ",
             "   4   2 "]

empty = replicate 9 "         "

